@@ -4,21 +4,11 @@ import numpy as np
 import os
 from numba import njit, prange
 
+from .maxcut_tfim_util import opencl_context
 
 IS_OPENCL_AVAILABLE = True
-ctx = None
-queue = None
 try:
     import pyopencl as cl
-    
-    # Pick a device (GPU if available)
-    ctx = cl.create_some_context()
-    queue = cl.CommandQueue(ctx)
-
-    # Load and build OpenCL kernels
-    kernel_src = open(os.path.dirname(os.path.abspath(__file__)) + "/kernels.cl").read()
-    program = cl.Program(ctx, kernel_src).build()
-    maxcut_hamming_cdf_kernel = program.maxcut_hamming_cdf
 except ImportError:
     IS_OPENCL_AVAILABLE = False
 
@@ -106,9 +96,28 @@ def maxcut_hamming_cdf(n_qubits, J_func, degrees, quality, hamming_prob):
     hamming_prob[-1] = 2.0
 
 
+@njit
+def binary_search(l, t):
+  left = 0
+  right = len(l) - 1
+
+  while left <= right:
+    mid = (left + right) >> 1
+
+    if l[mid] == t:
+      return mid
+
+    if l[mid] < t:
+      left = mid + 1
+    else:
+      right = mid - 1
+
+  return len(l)
+
+
 # Written by Elara (OpenAI custom GPT) and improved by Dan Strano
 @njit
-def local_repulsion_choice(adjacency_data, adjacency_vals, adjacency_rows, max_weight, weights, n, m):
+def local_repulsion_choice(G_cols, G_data, G_rows, max_weight, weights, n, m):
     """
     Pick m nodes out of n with repulsion bias:
     - High-degree nodes are already less likely
@@ -153,10 +162,19 @@ def local_repulsion_choice(adjacency_data, adjacency_vals, adjacency_rows, max_w
         mask[node] = True
 
         # Repulsion: penalize neighbors
-        for j in range(adjacency_rows[node], adjacency_rows[node + 1]):
-            nbr = adjacency_data[j]
+        for j in range(G_rows[node], G_rows[node + 1]):
+            nbr = G_cols[j]
             if available[nbr]:
-                weights[nbr] *= 0.5 ** (adjacency_vals[j] / max_weight)  # tunable penalty factor
+                weights[nbr] *= 0.5 ** (G_data[j] / max_weight)  # tunable penalty factor
+
+        for nbr in range(node):
+            if not available[nbr]:
+                continue
+            start = G_rows[nbr]
+            end = G_rows[nbr + 1]
+            j = binary_search(G_cols[start:end], node) + start
+            if j < end:
+                weights[nbr] *= 0.5 ** (G_data[j] / max_weight)  # tunable penalty factor
 
     return mask
 
@@ -168,8 +186,6 @@ def compute_energy(sample, G_data, G_rows, G_cols):
     for u in range(n_qubits):
         for col in range(G_rows[u], G_rows[u + 1]):
             v = G_cols[col]
-            if v < (u + 1):
-                continue
             energy += G_data[col] * (1 if sample[u] == sample[v] else -1)
 
     return energy
@@ -203,8 +219,6 @@ def sample_for_solution(G_data, G_rows, G_cols, shots, thresholds, J_eff):
     for u in range(n):
         for col in range(G_rows[u], G_rows[u + 1]):
             v = G_cols[col]
-            if v < (u + 1):
-                continue
             if best_solution[u] != best_solution[v]:
                 best_value += G_data[col]
 
@@ -217,13 +231,27 @@ def init_J_and_z(G_data, G_rows, G_cols):
     degrees = np.empty(n_qubits, dtype=np.uint32)
     J_eff = np.empty(n_qubits, dtype=np.float64)
     J_max = -float("inf")
-    for n in range(n_qubits):
-        start = G_rows[n]
-        end = G_rows[n + 1]
+    for r in range(n_qubits):
+        # Row sum
+        start = G_rows[r]
+        end = G_rows[r + 1]
         degree = end - start
-        J = -np.mean(G_data[start:end]) / degree if degree > 0 else 0
-        degrees[n] = degree
-        J_eff[n] = J
+        val = G_data[start:end].sum()
+
+        J = -val / degree if degree > 0 else 0
+        degrees[r] += degree
+        J_eff[r] += J
+
+        # Column sum
+        for idx in range(start, end):
+            c = G_cols[idx]
+            degrees[c] += 1
+            J_eff[c] += G_data[idx]
+
+    for r in range(n_qubits):
+        J = J_eff[r]
+        degree = degrees[r]
+        J_eff[r] = -J / degree if degree > 0 else 0
         J_abs = abs(J)
         if J_abs > J_max:
             J_max = J_abs
@@ -271,6 +299,18 @@ def init_theta(delta_t, tot_t, h_mult, n_qubits, J_eff, degrees):
     return theta
 
 
+def to_scipy_sparse_upper_triangular(G, nodes, n_nodes):
+    lil = lil_matrix((n_nodes, n_nodes), dtype=np.float64)
+    for u in range(n_nodes):
+        u_node = nodes[u]
+        for v in range(u + 1, n_nodes):
+            v_node = nodes[v]
+            if G.has_edge(u_node, v_node):
+                lil[u, v] = G[u_node][v_node].get('weight', 1.0)
+
+    return lil.tocsr()
+
+
 def maxcut_tfim_sparse(
     G,
     quality=None,
@@ -282,7 +322,7 @@ def maxcut_tfim_sparse(
     if isinstance(G, nx.Graph):
         nodes = list(G.nodes())
         n_qubits = len(nodes)
-        G_m = nx.to_scipy_sparse_array(G, weight='weight')
+        G_m = to_scipy_sparse_upper_triangular(G, nodes, n_qubits)
     else:
         n_qubits = G.shape[0]
         nodes = list(range(n_qubits))
@@ -307,7 +347,7 @@ def maxcut_tfim_sparse(
 
     if shots is None:
         # Number of measurement shots
-        shots = 1 << quality
+        shots = n_qubits << quality
 
     n_steps = 1 << quality
     grid_size = n_steps * n_qubits
@@ -333,20 +373,20 @@ def maxcut_tfim_sparse(
 
         # Move to GPU
         mf = cl.mem_flags
-        args_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=args)
-        J_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=J_eff)
-        deg_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=degrees)
-        theta_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=theta)
-        ham_buf = cl.Buffer(ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=hamming_prob)
+        args_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=args)
+        J_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=J_eff)
+        deg_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=degrees)
+        theta_buf = cl.Buffer(opencl_context.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=theta)
+        ham_buf = cl.Buffer(opencl_context.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=hamming_prob)
 
         # Kernel execution
-        maxcut_hamming_cdf_kernel(
-            queue, (grid_dim,), (group_size,),
+        opencl_context.maxcut_hamming_cdf_kernel(
+            opencl_context.queue, (grid_dim,), (group_size,),
             np.int32(n_qubits), deg_buf, args_buf, J_buf, theta_buf, ham_buf
         )
 
         # Fetch results
-        cl.enqueue_copy(queue, hamming_prob, ham_buf)
+        cl.enqueue_copy(opencl_context.queue, hamming_prob, ham_buf)
 
         hamming_prob /= hamming_prob.sum()
         tot_prob = 0.0
